/*
 * GkSu -- Gtk+ Frontend to su
 * Copyright (C) 2002 Gustavo Noronha Silva
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h> /* for basename() */
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <getopt.h>
#include <gtk/gtk.h>
#include <locale.h>

#include "../config.h"
#include "defines.h"

#include "gksu.h"

#include "util.h"
#include "gui.h"
#include "xauth.h"
#include "su.h"

struct option long_opts[] = {
    /*
     * { name  has_arg  *flag  val } 
     */
    {"help", no_argument, NULL, 'h'},
    {"login", no_argument, NULL, 'l'},
    {"preserv-env", no_argument, NULL, 'k'},
    {"user", required_argument, NULL, 'u'},
    {"print-pass", no_argument, NULL, 'p'},
    {"message", required_argument, NULL, 'm'},
    {"title", required_argument, NULL, 't'},
    {"icon", required_argument, NULL, 'i'},
    {"disable-grab", no_argument, NULL, 'g'},
    {"ssh-fwd", no_argument, NULL, 's'},
    {"debug", no_argument, NULL, 'd'},
    {0, 0, 0, 0}
};

/**
 * help:
 * @cmdname:  name of the command which was called by the user 
 * (argv[0])
 *
 * This function is a simple 'usage'-style printing function.
 * It is called if the user calls the program with --help or -h
 */
void
help (gchar *cmdname)
{ 
  g_print (_(
"GKsu version %s\n\n"
"Usage: %s [-u <user>] [-k] [-l] <command>\n\n"
"  --user <user>, -u <user>\n"
"    Calls <command> as the specified user\n"
"  --message <message>, -m <message>\n"
"    Replaces the standard message shown to ask for\n"
"    password for the argument passed to the option\n"
"  --title <title>, -t <title>\n"
"    Replaces the default title with the argument\n"
"  --icon <icon>, -i <icon>\n"
"    Replaces the default window icon with the argument\n"
"  --print-pass, -p\n"
"    Asks gksu to print the password to stdout, just\n"
"    like ssh-askpass. Useful to use in scripts with\n"
"    programs that accept receiving the passowrd on\n"
"    stdin.\n"
"  --disable-grab, -g\n"
"    Disables the \"locking\" of the keyboard, mouse,\n"
"    and focus done by the program when asking for\n"
"    password\n"
"  --ssh-fwd, -s\n"
"    Strip the host part of the $DISPLAY variable, so that\n"
"    GKSu will work on SSH X11 Forwarding.\n"
"\n" 
"  --login, -l\n"
"    Makes this a login shell. Beware this may cause\n"
"    problems with the Xauthority magic. Run xhost\n"
"    to allow the target user to open windows on your\n"
"    display!\n"
"  --preserv-env, -k\n"
"    Preserve the current environments, does not set $HOME\n"
"    nor $PATH, for example.\n"
"\n"
"\n"), PACKAGE_VERSION, cmdname);
}

/**
 * free_pass:
 * @pass: password to be freed securely
 *
 * Clears variables used to store password securely.
 */
void
free_pass (gchar *pass)
{
  if (pass != NULL)
    {
      memset (pass, 0, strlen(pass));
      g_free (pass);
      pass = NULL;
    }
}

int 
main (int argc, char **argv)
{
  gchar *myname;

  GkConf conf;

  int c;

  gboolean print_pass = FALSE;

  conf.xauth = NULL;
  conf.display = NULL;
  conf.dir_template = g_strdup ("/tmp/" PACKAGE_NAME "-XXXXXX");
  conf.dir= NULL;
  conf.user = NULL;
  conf.command = NULL; /* gksu */

  conf.title = NULL;
  conf.message = NULL;
  conf.icon = NULL;

  conf.login_shell = FALSE;
  conf.keep_env = FALSE;
  conf.grab = TRUE;

  conf.debug = FALSE;
  conf.ssh_fwd = FALSE;

  setlocale (LC_ALL, "");
  bindtextdomain(PACKAGE_NAME, LOCALEDIR);  
  bind_textdomain_codeset (PACKAGE_NAME, "UTF-8");
  textdomain(PACKAGE_NAME);
  gtk_set_locale();

  gtk_init (&argc, &argv);

  myname = basename (argv[0]);

  while ((c = getopt_long(argc, argv, "?hu:lpm:kt:i:gds", long_opts, NULL))
	 != EOF)
    {
      switch (c)
	{
	case 'h':
	  help (argv[0]);
	  exit(0);
	  break;
	case '?':
	  help (argv[0]);
	  exit(0);
	  break;
	case 'u':
	  conf.user = g_strdup (optarg);
	  break;
	case 'l':
	  conf.login_shell = TRUE;
	  break;
	case 'p':
	  print_pass = TRUE;
	  break;
	case 't':
	  conf.title = conv_to_utf8 (optarg);
	  break;
	case 'm':
	  conf.message = conv_to_utf8 (optarg);
	  break;
	case 'i':
	  conf.icon = g_strdup (optarg);
	  break;
	case 'k':
	  conf.keep_env = TRUE;
	  break;
	case 'g':
	  conf.grab = FALSE;
	  break;
	case 'd':
	  conf.debug = TRUE;
	  break;
	case 's':
	  conf.ssh_fwd = TRUE;
	  break;
	default:
	  fprintf (stderr, _("Using %s as username...\n"), argv[optind]);
	}
    }

  if (conf.title == NULL)
    conf.title = g_strdup (_("Changing user..."));

  if (conf.icon == NULL)
    gtk_window_set_default_icon_from_file (DATA_DIR "/pixmaps/gksu.png", 
					   NULL);
  else
    gtk_window_set_default_icon_from_file (conf.icon, NULL);

  /* if user is not specified, use root */
  if (conf.user == NULL)
    conf.user = g_strdup ("root");

  /* 
     if --print-pass is passed we do not need to care
     about xauth stuff and etc... let's just ask the
     password and print it
  */
  if (print_pass)
    {
      gchar *password;

      if (conf.message == NULL)
	password = ask_password (&conf,
				 _("Please enter %s's password"),
				 conf.user);
      else
	password = ask_password (&conf, conf.message);

      printf ("%s\n", password);
      free_pass (password);

      exit (0);
    }

  /* now we can begin to care about a command */
  if (argc == optind)
    {
      gk_dialog (GTK_MESSAGE_ERROR, _("Missing command to run."));
      exit (1);
    }

  if (prepare_xauth (&conf))
    return 1;
  conf.command = g_strdup (argv[optind]);

  su_do (&conf);
  
  g_free (conf.command);
  g_free (conf.display);

  g_free (conf.dir_template);
  g_free (conf.title);
  g_free (conf.user);

 return 0;
}
