/*
 * GkSu -- Gtk+ Frontend to su
 * Copyright (C) 2002 Gustavo Noronha Silva
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>

#include "defines.h"

#include "gksu.h"
#include "xauth.h"
#include "pixmaps.h"
#include "gui.h"
#include "util.h"

/**
 * ask_pass_ok_cb:
 * @w: activated widget
 * @data: the #GtkDialog to which a response should be emitted.
 *
 * This function handles cases in which the default widget
 * of the ask_pass dialog is not the one activated to 'OK'
 * the dialog. This function will, then, emmit the GTK_RESPONSE_OK
 * to the dialog passed on @data.
 */
void
ask_pass_ok_cb (GtkWidget *w, gpointer data)
{
  GtkWidget *dialog = (GtkWidget*)data;

  gtk_dialog_response (GTK_DIALOG(dialog),
		       GTK_RESPONSE_OK);
}

/**
 * ask_password:
 * @conf: the #GkConf filled with this run of gksu
 * @format: a printf()-like format that's going to be
 * set as the dialog's main label
 * @Varargs: variables that replace the codes on the format
 *
 * This is one of the main functions in gksu. It asks the
 * user for the password of the target user.
 *
 * Returns: a newly allocated gchar containing the password
 * or NULL if an error happens or the user cancels the action
 */
gchar*
ask_password (GkConf *conf, gchar *format, ...)
{
  gchar *pass = NULL;
	
  GtkWidget
    *dialog, 
    *hbox,
    *vbox,
    *image,
    *entry, 
    *label;

  gint response;

  va_list ap;
  gchar *msg;

  va_start(ap, format);
  msg = g_strdup_vprintf(format, ap);
  va_end(ap);

  /* dialog window */
  dialog = gtk_dialog_new_with_buttons (conf->title, NULL, GTK_DIALOG_MODAL,
					GTK_STOCK_CANCEL,
					GTK_RESPONSE_CANCEL,
					GTK_STOCK_OK,
					GTK_RESPONSE_OK,
					NULL);

  gtk_dialog_set_has_separator (GTK_DIALOG(dialog), FALSE);
  gtk_container_set_border_width (GTK_CONTAINER(dialog), 6);
  gtk_box_set_spacing (GTK_BOX(GTK_DIALOG(dialog)->vbox), 12);
  /* Center window and prepare for grab */
  gtk_window_set_default_size (GTK_WINDOW(dialog), 250, 100);
  gtk_window_set_position (GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
  gtk_window_stick (GTK_WINDOW(dialog));

  /* hbox */
  hbox = gtk_hbox_new (FALSE, 12);
  gtk_container_set_border_width (GTK_CONTAINER(hbox), 6);
  gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->vbox),
		      hbox, TRUE, TRUE, 0);
  gtk_widget_show (hbox);

  /* image */
  if (conf->icon == NULL)
    image = gtk_image_new_from_file (DATA_DIR"/pixmaps/gksu-auth.png");
  else
    image = gtk_image_new_from_file (conf->icon);
  gtk_misc_set_alignment (GTK_MISC(image), 0.5, 0);
  gtk_box_pack_start (GTK_BOX(hbox), image, FALSE, FALSE, 0);
  gtk_widget_show (image);

  /* vbox for label and entry */
  vbox = gtk_vbox_new (FALSE, 12);
  gtk_box_pack_start (GTK_BOX(hbox), vbox, TRUE, TRUE, 0);
  gtk_widget_show (vbox);

  /* label */
  label = gtk_label_new (msg);
  gtk_label_set_justify (GTK_LABEL(label), GTK_JUSTIFY_CENTER);
  gtk_label_set_use_markup (GTK_LABEL(label), TRUE);
  gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0);
  gtk_box_pack_start (GTK_BOX(vbox), label, TRUE, TRUE, 0);
  gtk_widget_show (label);
  g_free (msg);

  /* hbox for entry and label */
  hbox = gtk_hbox_new (FALSE, 6);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
  gtk_widget_show (hbox);

  /* entry label */
  label = gtk_label_new (_("Password:"));
  gtk_box_pack_start (GTK_BOX(hbox), label, FALSE, FALSE, 0);
  gtk_widget_show (label);
  
  /* entry */
  entry = gtk_entry_new();
  gtk_signal_connect (GTK_OBJECT(entry), "activate",
		      GTK_SIGNAL_FUNC(ask_pass_ok_cb),
		      dialog);
  gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
  gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
  gtk_widget_show (entry);
  
  gtk_widget_show_all (dialog);
  gtk_widget_grab_focus(entry);

  while (gtk_events_pending())
    gtk_main_iteration();

  if (conf->grab)
    {
      /* Grab focus */
      XGrabServer(GDK_DISPLAY());

      usleep (200); /* wait a bit longer for the grab, 
		       just in case system load is high */

      if (gdk_pointer_grab(dialog->window, TRUE, 0, NULL, NULL, 
			   GDK_CURRENT_TIME))
	{
	  gdk_pointer_ungrab(GDK_CURRENT_TIME);
	  XUngrabServer(GDK_DISPLAY());
	  gtk_widget_hide (dialog);
	  report_failed_grab ();
	  return NULL;
	}
      if (gdk_keyboard_grab(dialog->window, FALSE, GDK_CURRENT_TIME))
	{
	  gdk_pointer_ungrab(GDK_CURRENT_TIME);
	  XUngrabServer(GDK_DISPLAY());
	  gtk_widget_hide (dialog);
	  report_failed_grab ();
	  return NULL;
	}
    }

  /* let the magic begin */
  response = gtk_dialog_run (GTK_DIALOG(dialog));
  
  if (conf->grab)
    {
      /* Ungrab */
      XUngrabServer(GDK_DISPLAY());
      gdk_pointer_ungrab(GDK_CURRENT_TIME);
      gdk_keyboard_ungrab(GDK_CURRENT_TIME);
      gdk_flush();
    }

  /* Report passphrase if user selected OK */
  pass = g_strdup (gtk_entry_get_text(GTK_ENTRY(entry)));
		
  /* Kill the entry's copy of the passphrase. */
  gtk_entry_set_text(GTK_ENTRY(entry), "");
			
  switch (response)
    {
    case GTK_RESPONSE_CANCEL:
    case GTK_RESPONSE_DELETE_EVENT:
    case GTK_RESPONSE_NONE:
      free_pass (pass);
      clean_dir (conf->dir);
      exit (0);
      break;
    }

  gtk_widget_hide (dialog);

  while (gtk_events_pending())
    gtk_main_iteration();

  return pass;
}

/**
 * report_failed_grab:
 *
 * Simple utility function called to report a fail on
 * the process of grabbing the mouse or keyboard when
 * asking for the password through ask_password()
 */
void
report_failed_grab (void)
{
  gk_dialog (GTK_MESSAGE_ERROR, 
	     _("Could not grab keyboard or mouse.\n"
	       "A malicious client may be eavesdropping\n"
	       "on your session."));
}
